# protosfromdata.awk
#
# Thu Feb 21 09:54:54  2002
#
# selects a given number of prototypes per class from training data file
# as they come in the order of appearance, thus, data shuffling should
# be done before.
# used in trainer.awk
#
# Examples:
#   awk -f protosfromdata.awk x.dat
#   selects *one* prototype per class from file x.dat
#
#   awk -vNO=3 -f protosfromdata.awk - 
#   selects three prototypes per class from standard input (probably pipe mode)
#
#   awk -vNO='0:2,1:3,2:1' -f protosfromdata.awk x.dat
#   selects 2,3,1 prototypes from x.dat for class 0,1,2, respectively.
# 


BEGIN {

  if(NO == "") 
  
    NO = 1
  
  else {
  
    nf = split(NO, noprotosperclass, ",")
    
    i = 1

    if(nf == 1) {
      if(NO ~ /[^0-9]/)
        x("no_protos_per_class must be integer")
      
      i = nf + 1
    }

    for(i = i; i <= nf; i++) {

      if(split(noprotosperclass[i], nppc, ":") != 2)
        x("C1:#1,C2:#2,...,Ck:#k format needed")
        
      noprotos[nppc[1]] = nppc[2]
      
      NO = 0
    }
  }
}

{
  if($0 ~ /^[^#]/) {
    if((NO > 0 && cnt[$NF] < NO) || cnt[$NF] < noprotos[$NF]) {
      ++cnt[$NF]
      print 
    }
  }
}

END {

  for(c in cnt) 
    if(NO <= 0 && noprotos[c] == 0)
      x("classes without codebooks!")
    else if((NO>0 && cnt[c] < NO) || cnt[c] < noprotos[c])
      x("not enough codebooks generated!")

}

function x(txt) {
  print txt | "cat >&2"
  exit(1)
}
