# trainer.awk
#
# Thu Feb 21 07:32:35  2002
# 
# external commands used: srng, gzip, cat, test, awk (must be in PATH, also .)
# external scripts used: shuffle.awk, protosfromdata.awk (must be in cwd)
# internal shell command: if
#
# This awk has been tested for linux and cygwin installations.
# Check the BEGIN section below.


BEGIN {

 # srand is to slow: 1 change/s only (for cygwin)
  srnd()

 # redirection to standard error channel
  STDERR = "cat >&2"
 
 # testmode to be controled via (g)awk -vTEST=1 -f trainer.awk ... 
  TEST = (TEST != "") ? 1 : 0
 
 # jobcounter
  jobs = 0 
  
 # reinit for each encountered job
  reinit = 1

 # how to call the shuffle script
  shuffle = "awk -vSRAND=0 -f shuffle.awk -"
 
 # filter empty lines and lines *starting* with a comment operator #
  filter = "awk '!/(^#)|(^[ \t]*$)/' -"

}


reinit==1 {
  paramcnt["train"] = paramcnt["test"] = paramcnt["parameter"] = 0
  paramcnt["protos"] = paramcnt["lambdas"] = paramcnt["lambdaweights"] = 0
  paramcnt["outfile"] = paramcnt["repeat"] = paramcnt["cycles"] = 0
  paramcnt["validatesteps"] = 0
  reinit = 0
}


!/(^#)|(^[\t\r ]*$)/ {

  gsub(/\t/,"")

  if(split($0,assign,"=") != 2) 
    {x("multi- and nonassigment not allowed!"); next}

  gsub(/^[\t ]+|[\t\r ]+$/,"",assign[1])
  gsub(/^[\t ]+|[\t\r ]+$/,"",assign[2])


  param = tolower(assign[1])

  if(!(param in paramcnt)) {x("OOPS unknown command >" param "< !"); next} 
  if(++paramcnt[param]>1) x("Command appears multi, overwriting old value!")

  parameter[param] = assign[2]
  
  todo = 0
  for(param in paramcnt) if(paramcnt[param] == 0) todo = 1


  if(todo == 0) {

    reinit = 1
    ++jobs
    

    repeat = parameter["repeat"]


    cycles = parameter["cycles"]
    if(cycles <= 0)
      {x("cycle number must be > 0"); next}


    validatesteps = parameter["validatesteps"]
    if(validatesteps <= 0)
      {x("validatesteps number must be > 0"); next}
    

    trainfile = parameter["train"]
    
    if(system("test -f '" trainfile "'") != 0) 
      {x("no such train-datafile: " trainfile); next}
    
    
    testfile = parameter["test"]
    partition = (testfile ~ /[a-zA-Z]/) ? -1. : testfile + 0.
    
    if(partition < 0) {
      if(system("test -f '" testfile "'") != 0)
        {x("no such test-datafile: " testfile); next}
    }
    else if(partition < 0 || partition > 1) 
      {x("invalid test data partition interval: " partition); next}


    infilecompress = system("gzip -t " trainfile)
    
    catrainfile = "" ((infilecompress == 0) ? "gzip -d -c " : "cat ") trainfile

    validate = catrainfile "|" filter "|"
    validate = validate "awk '{if(++arr[$NF] == 1) ++c} END{print c, NF, NR}' -"
    validate | getline
    close(validate)
    
    noclasses    = $1 + 0
    nodimensions = $2 - 1
    nopatterns   = $3 + 0

    
    print "======= Job " jobs " ======="
    print "cycles: " cycles
    print "validation every: " validatesteps
    print "" (infilecompress ? "" : "compressed ") "training data: " trainfile
    print "number of patterns: " nopatterns
    print "number of classes: " noclasses
    print "noclasses of dimensions: " nodimensions
    print "repeats: " repeat


    for(rep = 1; rep <= repeat; rep++) {


      if(partition >= 0) {
      
        srndpart = srnd()
        sub(/[0-9]+/, srndpart, shuffle)
       
        thresh = int(nopatterns*partition) 
        cmd_gettest = catrainfile "|" filter "|" shuffle "|"
        cmd_gettest = cmd_gettest "awk '{if(NR>" thresh ") exit; print}' -"
        cmd_gettest = cmd_gettest "|awk 'BEGIN{print \"#" thresh " " nodimensions"\"}{print}' -"

        cmd_gettrain = cmd_gettest
        sub(/exit; /, "", cmd_gettrain)
        sub(/#[0-9]+/,"#" (nopatterns-thresh), cmd_gettrain)
   
        print rep ") srand(" srndpart ")om partition-split (" 100 * partition "%)"
      }
      else {

        cmd_gettrain = catrainfile "|" filter 
        cmd_gettrain = cmd_gettrain "|awk 'BEGIN{print \"#" nopatterns " " nodimensions"\"}{print}' -"
        cmd_gettest = "cat " testfile

        print rep ") using test data: " testfile
      }

      noprotos = 0
      prototypes[0] = ""


      parameters = parameter["parameter"]
      parametersatfile = sub(/^@/, "", parameters)
      
      if(length(parameters) < 255 && system("test -f '" parameters "'") == 0) {
      
        print "" rep ") using parameter file" \
          ((parametersatfile>0) ? " (@mode): " : ": ") "" parameters
        
        cmd = "cat " parameters

        while((cmd | getline parameters) > 0)
          if(!(parameters ~ /(^#)|(^[\r\t ]*$)/)) break
          
        close(cmd)

        gsub(/^[\t ]+|[\t\r ]+$/, "", parameters)

        if(parameters == "") {x("problem with: " cmd); break}
        
        if(parametersatfile > 0) 
          sub(/^[0-9]+ /, "", parameters)
      }
      if(split(parameters, dummy, " ") != 7) 
        {x("wrong number of parameters"); break}
      
      
      protos = parameter["protos"]
      protosatfile = sub(/^@/, "", protos)
      
      if(system("test -f '" protos "'") != 0) {
        
        if(protos ~ /[^0-9]/) 
          {x("protos must be integer or valid file name"); break}

        srndprotos = srnd()
        sub(/[0-9]+/, srndprotos, shuffle)

        cmd_getprotos = catrainfile "|" shuffle " | awk -vNO='"
        cmd_getprotos = cmd_getprotos protos "' -f protosfromdata.awk -"

        print "" rep ") srand(" srndprotos ")om prototype selection"
      }
      else {

        cmd_getprotos = "cat " protos "|" filter
      
        print "" rep ") using prototype file" \
          ((protosatfile>0) ? " (@mode): " : ": ") "" protos

        if(protosatfile > 0) noprotos = -1
      }


      while ((cmd_getprotos | getline) > 0 && !($0 ~ /(^#)|(^[\r\t ]*$)/)) {
        gsub(/^[\t ]+|[\t\r ]+$/,"")
        prototypes[++noprotos] = $0
      }
      close(cmd_getprotos)

      while(split(prototypes[noprotos], dummy, " ") != nodimensions + 1)
        if(--noprotos <= 0) break;
        
      if(noprotos <= 0) {x("Could not read protos: " protos); break}


      lambdas = parameter["lambdas"]
      lambdasatfile = sub(/^@/, "", lambdas)
      
      if(length(lambdas) < 255 && system("test -f '" lambdas "'") == 0) {
      
        print "" rep ") using lambdas file" \
          ((lambdasatfile>0) ? " (@mode): " : ": ") "" lambdas
        
        cmd = "cat " lambdas

        if(lambdasatfile == 0) {
        
          while((cmd | getline lambdas) > 0)
            if(!(lambdas ~ /(^#)|(^[\r\t ]*$)/)) break

          gsub(/^[\t ]+|[\t\r ]+$/, lambdas)

          if(lambdas == "") {x("problem with: " cmd); break}
          
        }
        else {
        
          recent = last = ""
          while((cmd | getline) > 0 && !($0 ~ /(^#)|(^[\r\t ]*$)/)) {
            recent = last
            gsub(/^[\t ]+|[\t\r ]+$/,"")
            last = $0
          }

          if(recent != "" && split(recent, dummy, " ") < nodimensions + 1) 
            lambdas = recent
          else if(last != "" && split(last, dummy, " ") < nodimensions + 1) 
            lambdas = last
          else
            lambdas = "-1"
        }
        close(cmd)
      }
      if(lambdas+0. != -1 && (tmp = split(lambdas, dummy, " ")) != nodimensions)
        {x("number of lambdas must be " nodimensions " not " tmp); break}


      lambdaweights = parameter["lambdaweights"]
      lambdaweightsatfile = sub(/^@/, "", lambdaweights)
      
      if(length(lambdaweights) < 255 && system("test -f '" lambdaweights "'") == 0) {
      
        print "" rep ") using lambdaweights file" \
          ((lambdaweightsatfile>0) ? " (@mode): " : ": ") "" lambdaweights
        
        cmd = "cat " lambdaweights

        if(lambdaweightsatfile == 0) { 
          
          while((cmd | getline lambdaweights) > 0)
            if(!(lambdaweights ~ /(^#)|(^[\r\t ]*$)/)) break

          gsub(/^[\t ]+|[\t\r ]+$/, lambdaweights)

          if(lambdaweights == "") {x("problem with: " cmd); break}
        }
        else {
        
          recent = last = lambdaweights = ""
          while((cmd | getline) > 0 && !($0 ~ /(^#)|(^[\t\r ]*$)/)) {
            recent = last
            gsub(/^[\t ]+|[\t\r ]+$/,"")
            last = $0
          }

          if(recent != "" && last != "" && \
              split(recent, dummy, " ") < nodimensions + 1 && \
              split(last, dummy, " " ) < nodimensions + 1) 
            lambdaweights = last
        }
        close(cmd)
      }
      if(lambdaweights != "" && (tmp = split(lambdaweights, dummy, " ")) != nodimensions)
        {x("number of lambdaweights must be " nodimensions " not " tmp); break}

      
      print rep ") parameters: " parameters


      prototypes[0] = noprotos " " parameters
      prototypes[++noprotos] = lambdas
      prototypes[++noprotos] = lambdaweights


      outfile = parameter["outfile"]
      if(outfile == "") {x("empty outfile names not allowed"); break}

      outfileoverwrite = gsub(/^!/, "", outfile)
      outfilecompress = gsub(/^-/, "", outfile)
      gsub(/%/, rep, outfile)
      gsub(/\$/, jobs, outfile)
      
      if(outfilecompress > 0) outfile = outfile ".gz"
      
      if(outfileoverwrite == 0) {
      
        nf = sub(/\.[^\.]*$/,".!&",outfile)
        
        if(nf==0) outfile = outfile ".!"
        
        cnt = 0
        
        do {
          fakename = outfile
          sub(/.!/, ((cnt++ == 0) ? "" : "" cnt), fakename)
        } while(system("test -f '" fakename "'") == 0)

        outfile = fakename
      }
      print "" rep ") " ((outfileoverwrite>0) ? "over" :"") "writing" \
        ((outfilecompress>0) ? " compressed" : "") \
        " output file: " outfile
      
      tmpinit = "tmpXxx" srnd()
      system("rm -f " tmpinit)
      
      for(i = 0; i <= noprotos; i++)
        print prototypes[i] > tmpinit
      close(tmpinit)

      
      cmd_train = cmd_gettrain "| srng " tmpinit \
        ((TEST == 0) ? (" " cycles " " validatesteps) : " 0") "" \
        ((outfilecompress>0) ? "| gzip -9 -c" : "") " > " outfile 

# print cmd_train

      print "Training. Job " jobs " #" rep " > " outfile "..." | STDERR

      if(system(cmd_train) != 0) 
        {x("error invoking training algorithm with file: " tmpinit); break}

      system("rm -f " tmpinit)

      print "   ...training done." | STDERR


      cmd_valid = "(" cmd_gettest " && " ((outfilecompress) ? \
        "gzip -d -c ": "cat ") outfile \
        ((TEST == 0) ? "" : "| awk '/^$/ {p=1} p==1' -")  "" \
        ") | srng - 0 | awk '/^-1 /{$1=\"\"; print}' > " outfile ".test"

# print cmd_valid

      print "Testing. Job " jobs " #" rep " > " outfile ".test ..." | STDERR

      if(system(cmd_valid) != 0) 
        {x("error invoking training algorithm for testing:" cmd_valid); break}

      system("cat " outfile ".test |" STDERR)

      print "   ...testing done." | STDERR

    } #for rep
    
    if(rep == repeat + 1 || repeat <= 0)
      ++successjobs
    else 
      print "Error: Job" jobs
  }
}


END {

  print "" (successjobs+0) " job(s) done."
  
  if(successjobs != jobs || todo == 1) {
    print "Batchfile incomplete or truncated or erroneous!" | STDERR
    exit(1)
  }
}



function x(txt) {
  print "job " jobs " < L" NR ": " txt " Skipping job." | STDERR
}


function srnd() {
  while(xsrnd <= 1) xsrnd = srand();
  return int(rand()*2*xsrnd)
}

# EOF
